import asyncio
import threading

from typing import Awaitable, TypeVar

import pandas as pd

from intugle.core import settings
from intugle.core.llms.embeddings import Embeddings
from intugle.core.semantic_search.crud import SemanticSearchCRUD
from intugle.core.semantic_search.semantic_search import HybridDenseLateSearch
from intugle.core.utilities.processing import string_standardization
from intugle.parser.manifest import ManifestLoader

T = TypeVar("T")


def _run_async_in_sync(coro: Awaitable[T]) -> T:
    """
    Runs an async coroutine in a sync context, handling cases where an event loop is already running.
    """
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(coro)

    if loop.is_running():
        result = None
        exc = None

        def thread_target():
            nonlocal result, exc
            try:
                result = asyncio.run(coro)
            except Exception as e:
                exc = e

        thread = threading.Thread(target=thread_target)
        thread.start()
        thread.join()

        if exc:
            raise exc
        return result
    else:
        return loop.run_until_complete(coro)


class SemanticSearch:
    def __init__(
        self, project_base: str = settings.PROJECT_BASE, collection_name: str = settings.VECTOR_COLLECTION_NAME
    ):
        self.manifest_loader = ManifestLoader(project_base)
        self.manifest_loader.load()
        self.manifest = self.manifest_loader.manifest
        self.collection_name = collection_name

        self.project_base = project_base

    def get_column_details(self):
        sources = self.manifest.sources
        models = self.manifest.models

        column_details = []
        for source in sources.values():
            table = source.table
            for column in table.columns:
                metrics = column.profiling_metrics.model_dump()
                count = metrics.get("count", 0)
                distinct_count = metrics.get("distinct_count", 0)
                null_count = metrics.get("null_count", 0)

                uniqueness = distinct_count / count if count > 0 else 0
                completeness = (count - null_count) / count if count > 0 else 0

                column_detail = {
                    "id": f"{table.name}.{column.name}",
                    "column_name": column.name,
                    "column_glossary": column.description,
                    "column_tags": column.tags,
                    "category": column.category,
                    "table_name": table.name,
                    "table_glossary": table.description,
                    "uniqueness": uniqueness,
                    "completeness": completeness,
                    **metrics,
                }
                column_details.append(column_detail)

        for model in models.values():
            for column in model.columns:
                metrics = column.profiling_metrics.model_dump()
                count = metrics.get("count", 0)
                distinct_count = metrics.get("distinct_count", 0)
                null_count = metrics.get("null_count", 0)

                uniqueness = distinct_count / count if count > 0 else 0
                completeness = (count - null_count) / count if count > 0 else 0

                column_detail = {
                    "id": f"{table.name}.{column.name}",
                    "column_name": column.name,
                    "column_glossary": column.description,
                    "column_tags": column.tags,
                    "category": column.category,
                    "table_name": table.name,
                    "table_glossary": table.description,
                    "uniqueness": uniqueness,
                    "completeness": completeness,
                    **metrics,
                }
                column_details.append(column_detail)

        return column_details

    async def _async_initialize(self):
        embeddings = Embeddings(settings.EMBEDDING_MODEL_NAME, settings.TOKENIZER_MODEL_NAME)
        semantic_search_crud = SemanticSearchCRUD(self.collection_name, [embeddings])
        column_details = self.get_column_details()
        column_details = pd.DataFrame.from_records(column_details)
        await semantic_search_crud.initialize(column_details)

    def initialize(self):
        return _run_async_in_sync(self._async_initialize())

    async def _search_async(self, query):
        embeddings = Embeddings(settings.EMBEDDING_MODEL_NAME, settings.TOKENIZER_MODEL_NAME)
        semantic_search = HybridDenseLateSearch(self.collection_name, embeddings)

        data = await semantic_search.search(string_standardization(query))
        return data

    def search(self, query):
        search_results = _run_async_in_sync(self._search_async(query))
        if search_results.shape[0] == 0:
            return search_results
        search_results.sort_values(by="score", ascending=False, inplace=True)

        column_details = self.get_column_details()
        column_details_df = pd.DataFrame.from_records(column_details)
        merged_df = pd.merge(
            search_results, column_details_df, left_on="column_id", right_on="id", how="left"
        ).drop(columns=["id"])
        return merged_df


import time

from typing import Any, Optional

import duckdb
import numpy as np
import pandas as pd

from intugle.adapters.adapter import Adapter
from intugle.adapters.factory import AdapterFactory
from intugle.adapters.models import (
    ColumnProfile,
    DataSetData,
    ProfilingOutput,
)
from intugle.adapters.types.duckdb.models import DuckdbConfig
from intugle.adapters.utils import convert_to_native
from intugle.common.exception import errors
from intugle.core.utilities.processing import string_standardization


class DuckdbAdapter(Adapter):

    def __init__(self):
        duckdb.install_extension('httpfs')
        duckdb.load_extension('httpfs')        

    @staticmethod
    def check_data(data: Any) -> DuckdbConfig:
        try:
            data = DuckdbConfig.model_validate(data)
        except Exception:
            raise TypeError("Input must be a duckdb config.")
        return data

    def profile(self, data: DuckdbConfig, table_name: str) -> ProfilingOutput:
        """
        Generates a profile of a file.

        Args:
            df: The input pandas DataFrame.

        Returns:
            A pydantic model containing the profile information:
            - "count": Total number of rows.
            - "columns": List of column names.
            - "dtypes": A dictionary mapping column names to generalized data types.
        """
        data = self.check_data(data)

        def __format_dtype__(dtype: Any) -> str:
            """Maps dtype to a generalized type string."""
            type_map = {
                "VARCHAR": "string",
                "DATE": "date & time",
                "BIGINT": "integer",
                "DOUBLE": "float",
                "FLOAT": "float",
            }
            return type_map.get(dtype, "string")

        self.load(data, table_name)

        # Fetching total count of table
        query = f"""
        SELECT count(*) as count FROM "{table_name}"
        """
        data = duckdb.execute(query).fetchall()

        total_count = data[0][0]

        # Fetching column name and their data types of table
        query = """
        SELECT column_name, data_type
        FROM duckdb_columns()
        WHERE table_name = ?
        """
        data = duckdb.execute(query, [table_name]).fetchall()

        dtypes = {col: __format_dtype__(dtype) for col, dtype in data}
        columns = [col for col, _ in data]

        return ProfilingOutput(
            count=total_count,
            columns=columns,
            dtypes=dtypes,
        )

    def column_profile(
        self,
        data: DuckdbConfig,
        table_name: str,
        column_name: str,
        total_count: int,
        sample_limit: int = 10,
        dtype_sample_limit: int = 10000,
    ) -> Optional[ColumnProfile]:
        """
        Generates a detailed profile for a single column of a pandas DataFrame.

        It calculates null and distinct counts, and generates two types of samples:
        1.  `sample_data`: A sample of unique values.
        2.  `dtype_sample`: A potentially larger sample combining unique values with
            random non-unique values, intended for data type analysis.

        Args:
            df: The input pandas DataFrame.
            column_name: The name of the column to profile.
            sample_limit: The desired number of items for the data samples.

        Returns:
            A dictionary containing the profile for the column, or None if the
            column does not exist.
        """
        data = self.check_data(data)

        self.load(data, table_name)

        start_ts = time.time()

        # --- Calculations --- #
        query = f"""
        SELECT 
            COUNT(DISTINCT "{column_name}") AS distinct_count,
            SUM(CASE WHEN "{column_name}" IS NULL THEN 1 ELSE 0 END) AS null_count
        FROM  
            "{table_name}"
        """
        distinct_null_data = duckdb.execute(query).fetchall()

        distinct_count, null_count = distinct_null_data[0]
        not_null_count = total_count - null_count

        # --- Sampling Logic --- #
        # 1. Get a sample of distinct values.
        sample_query = f"""
        SELECT 
            DISTINCT CAST( "{column_name}" AS VARCHAR) AS sample_values 
        FROM 
            "{table_name}"
        WHERE 
            "{column_name}" IS NOT NULL LIMIT {dtype_sample_limit}
        """
        data = duckdb.execute(sample_query).fetchall()
        distinct_values = [d[0] for d in data]

        not_null_series = pd.Series(distinct_values)

        if distinct_count > 0:
            distinct_sample_size = min(distinct_count, dtype_sample_limit)
            sample_data = list(np.random.choice(distinct_values, distinct_sample_size, replace=False))
        else:
            sample_data = []

        # 2. Create a combined sample for data type analysis.
        dtype_sample = None
        if distinct_count >= dtype_sample_limit:
            # If we have enough distinct values, that's the best sample.
            dtype_sample = sample_data
        elif distinct_count > 0 and not_null_count > 0:
            # If distinct values are few, supplement them with random non-distinct values.
            remaining_sample_size = dtype_sample_limit - distinct_count

            # Use replace=True in case the number of non-null values is less than the remaining sample size needed.
            additional_samples = list(not_null_series.sample(n=remaining_sample_size, replace=True))

            # Combine the full set of unique values with the additional random samples.
            dtype_sample = list(distinct_values) + additional_samples
        else:
            dtype_sample = []

        # --- Convert numpy types to native Python types for JSON compatibility --- #
        native_sample_data = convert_to_native(sample_data)
        native_dtype_sample = convert_to_native(dtype_sample)

        business_name = string_standardization(column_name)

        # --- Final Profile --- #
        return ColumnProfile(
            column_name=column_name,
            business_name=business_name,
            table_name=table_name,
            null_count=null_count,
            count=total_count,
            distinct_count=distinct_count,
            uniqueness=distinct_count / total_count if total_count > 0 else 0.0,
            completeness=not_null_count / total_count if total_count > 0 else 0.0,
            sample_data=native_sample_data[:sample_limit],
            dtype_sample=native_dtype_sample,
            ts=time.time() - start_ts,
        )

    @staticmethod
    def _get_load_func(data: DuckdbConfig):
        func = {
            "csv": "read_csv",
            "parquet": "read_parquet",
            "xlsx": "read_xlsx",
        }
        ld_func = func.get(data.type)
        if ld_func is None:
            raise errors.NotFoundError(f"Type: {data.type} not supported")

        if data.type == "xlsx":
            return f"{ld_func}('{data.path}', ignore_errors = true)"  # Replace cells that can't be cast to the corresponding inferred column type with NULL

        return f"{ld_func}('{data.path}')"

    def load_view(self, data: DuckdbConfig, table_name: str):
        query = f"""CREATE OR REPLACE VIEW "{table_name}" AS {data.path}"""
        duckdb.execute(query)

    def load_file(self, data: DuckdbConfig, table_name: str):
        ld_func = self._get_load_func(data)
        query = f"""CREATE VIEW IF NOT EXISTS "{table_name}" AS SELECT * FROM {ld_func};"""

        duckdb.execute(query)

    def load(self, data: DuckdbConfig, table_name: str):
        data = self.check_data(data)
        if data.type == "query":
            self.load_view(data, table_name)
            return
        if data.type == "table":
            # The table is already materialized in DuckDB, so there's nothing to load.
            return
        self.load_file(data, table_name)
    
    def execute_df(self, query):
        df = duckdb.sql(query).to_df()
        return df

    def to_df(self, _: DuckdbConfig, table_name: str):
        query = f'''SELECT * from "{table_name}"'''
        df = self.execute_df(query)
        return df

    def to_df_from_query(self, query: str) -> pd.DataFrame:
        return duckdb.sql(query).to_df()

    def create_table_from_query(self, table_name: str, query: str):
        duckdb.sql(f'CREATE OR REPLACE VIEW "{table_name}" AS {query}')

    def create_new_config_from_etl(self, etl_name: str) -> "DataSetData":
        return DuckdbConfig(path=etl_name, type="table")

    def deploy_semantic_model(self, semantic_model_dict: dict, **kwargs):
        """Deploys a semantic model to the target system."""
        raise NotImplementedError("Deployment is not supported for the DuckdbAdapter.")

    def execute(self, query):
        df = self.execute_df(query)
        data = df.to_dict(orient="records")
        return data

    async def aexecute(self, query):
        result = duckdb.sql(query).fetchnumpy()
        df = pd.DataFrame(result)
        data = df.to_dict(orient="records")
        return data

    def intersect_count(self, table1: "DataSet", column1_name: str, table2: "DataSet", column2_name: str) -> int:
        table1_name = table1.name
        table2_name = table2.name
        query = f"""
        SELECT COUNT(*) as intersect_count FROM (
            SELECT DISTINCT "{column1_name}" FROM "{table1_name}" WHERE "{column1_name}" IS NOT NULL
            INTERSECT
            SELECT DISTINCT "{column2_name}" FROM "{table2_name}" WHERE "{column2_name}" IS NOT NULL
        ) as t
        """
        result = self.execute(query)
        return result[0]['intersect_count']

    def get_details(self, data: DuckdbConfig):
        data = self.check_data(data)
        return data.model_dump()


def can_handle_pandas(df: Any) -> bool:
    try:
        df = DuckdbAdapter.check_data(df)
    except Exception:
        return False
    return True


def register(factory: AdapterFactory):
    factory.register("duckdb", can_handle_pandas, DuckdbAdapter)



from copy import deepcopy

from ..models.models import FieldDetailsModel, FilterModel, WildCardOptions

STRING_OPERATOR = [
    WildCardOptions.starts_with,
    WildCardOptions.ends_with,
    WildCardOptions.contains,
    WildCardOptions.exactly_matches,
]

INITIAL_FILTERS = {"fields": [], "condition": "AND"}


class Filter:
    def __init__(
        self,
        filter: FilterModel,
        field_details: dict[int | str, FieldDetailsModel],
        can_view_pii: bool = True,
    ):
        self._filters = deepcopy(INITIAL_FILTERS)
        self._filter = filter
        self._field_details = field_details
        self._can_view_pii = can_view_pii
        self._assets = set()

    def _get_wildcard_value(self, option: WildCardOptions, value):
        if option == WildCardOptions.starts_with:
            return f"{value}%"
        if option == WildCardOptions.ends_with:
            return f"%{value}"
        if option == WildCardOptions.contains:
            return f"%{value}%"
        if option == WildCardOptions.exactly_matches:
            return value
        return value

    def _get_wildcard_operator(self, option: WildCardOptions, exclude):
        # if option == WildCardOptions.starts_with:
        #     return "like"
        if option in STRING_OPERATOR:
            if exclude:
                return "not_like"
            return "like"
        return option

    def build_selection(self):
        _filter = self._filter
        if _filter is not None and _filter.selections:
            selections = []
            for selection in _filter.selections:
                field = self._field_details[selection.id]
                fields = []

                self._assets.add(field.asset_id)

                if selection.values:
                    sel = {
                        "dataset": field.asset_name,
                        "field": field.name,
                        "sql_code": field.sql_code,
                        "operator": "not_in" if selection.exclude else "in",
                        "md5_hash": field.is_pii and not self._can_view_pii,
                        "params": {"value": selection.values, "type": "list"},
                    }
                    if selection.dimFunc:
                        sel["isFunction"] = True
                        sel["function"] = [{"funName": selection.dimFunc, "params": {"expr": field.sql_code}}]
                    fields.append(sel)

                if selection.null:
                    sel_null = {
                        "dataset": field.asset_name,
                        "field": field.name,
                        "sql_code": field.sql_code,
                        "operator": "is_not_null" if selection.exclude else "is_null",
                        "md5_hash": field.is_pii and not self._can_view_pii,
                        "params": {"value": "", "type": "list"},
                    }
                    if selection.dimFunc:
                        sel_null["isFunction"] = True
                        sel_null["function"] = [{"funName": selection.dimFunc, "params": {"expr": field.sql_code}}]
                    fields.append(sel_null)

                selections.append({
                    "fields": fields,
                    "condition": "AND" if selection.exclude else "OR",
                })

            self._filters["fields"].extend(selections)

    def build_wildcard(self):
        _filter = self._filter
        if _filter is not None and _filter.wildcards:
            wildcards = []
            for wildcard in _filter.wildcards:
                field = self._field_details[wildcard.id]

                self._assets.add(field.asset_id)

                value = self._get_wildcard_value(wildcard.option, wildcard.value)
                operator = self._get_wildcard_operator(wildcard.option, wildcard.exclude)
                wc = {
                    "dataset": field.asset_name,
                    "field": field.name,
                    "sql_code": field.sql_code,
                    "operator": operator,
                    "md5_hash": field.is_pii and not self._can_view_pii,
                    "params": {"value": value},
                }
                if wildcard.dimFunc:
                    wc["isFunction"] = True
                    wc["function"] = [
                        {
                            "funName": wildcard.dimFunc,
                            "params": {"expr": field.sql_code},
                        }
                    ]
                wildcards.append(wc)

            self._filters["fields"].extend(wildcards)

    def get_filter(self):
        return self._filters

    def build_filters(self):
        if self._filter is not None:
            self.build_selection()
            self.build_wildcard()

    def destroy_filters(self):
        self._filters = deepcopy(INITIAL_FILTERS)

    def get_fresh_filter(self):
        self.destroy_filters()
        self.build_filters()
        return self.get_filter()

    def get_assets(self):
        return self._assets



from enum import Enum

from pydantic import BaseModel, ConfigDict


class SchemaBase(BaseModel):
    """Base model configuration"""

    model_config = ConfigDict(
        extra="ignore",
        use_enum_values=True,
        validate_by_name=True
    )
    

class NodeType(str, Enum):
    MODEL = "model"
    RELATIONSHIP = "relationship"
    FEWSHOT = "fewshot"
    ANALYTICS_CATALOGUE = "analytics_catalogue"
    SOURCE = "source"





import logging
import time

from enum import Enum
from typing import List, Optional, Tuple

import pandas as pd

from langchain_core.messages import HumanMessage
from langchain_core.output_parsers import JsonOutputParser
from langchain_core.prompts import PromptTemplate
from langgraph.errors import GraphRecursionError
from tqdm import tqdm

from intugle.analysis.models import DataSet
from intugle.core import settings
from intugle.core.llms.chat import ChatModelLLM
from intugle.core.pipeline.link_prediction.utils import (
    dtype_check,
    extract_innermost_dict,
    linkage,
    preprocess_profiling_df,
)
from intugle.core.utilities.llm_utils import (
    generate_create_table_query,
    read_column_datatypes,
)

log = logging.getLogger(__name__)


class Status(str, Enum):
    HALLUCINATED_INVOKE = "llm_hallicunated_invoke"
    HALLUCINATED_RECURSION = "llm_hallicunated_recursion"
    COMPLETED = "completed"
    NO_LINKS = "no_links"


# Add langfuse and openai handler


class LinkPredictionAgentic:
    HALLUCINATIONS_MAX_RETRY = settings.HALLUCINATIONS_MAX_RETRY

    PROMPT_TEMPLATE = """<|begin_of_text|><|start_header_id|>system<|end_header_id|> Task: Identifying Foreign Keys Between Two Tables
You are tasked with identifying potential foreign key relationships between two tables based on the following criteria:
Context: You are in attempt {current_attempt}. Please consider the conditions outlined below before making a determination.

Conditions for Identifying Foreign Keys:
- Uniqueness Requirement: If a single pair of foreign keys are identified then uniqueness for atleast one of columns should be greater than 80%.
- Semantic Relationship: The column names in the two tables should have a logical and semantic relationship, or there should be logical and semantic relationship between column name and table names of links. For example, a column named order_id in the child table likely corresponds to order_id or a similar identifier in the parent table, and also column named order_id in shipping table is likely to correspond to a table named orders.
- Data Format and Data Type Consistency: The data format and pattern of values in the potential foreign key column should match the referenced column in the parent table. Ensure that the data types align (e.g., integers, alphanumeric) and that there is no significant mismatch in data format.
- created_date, modified_date or updated_date or any other columns indicating data entry updates should not be considered as foreign key.

Steps:
In attempts 1-3, you will attempt to identify foreign keys involving a single column in each table.
If, based on the provided metadata, no link is identified between the two tables, return 'NA' for column and table names

Input Structure:
Metadata Information for the Tables:
The metadata for each table will include:
count_distinct: The number of unique values in a column.
uniqueness: The percentage of distinct values to the total number of rows.
completeness: The percentage of non-null entries in the column.
datatype: The data type of the column (e.g., integer-Dimension, alphanumeric).
sample data: A few example values from the column.

{format_instructions}


If no valid link is identified, return:
'NA' for column and table names

Input: {previous_attempts} {table_info}

Based on your evaluation of the data and metadata, please proceed to attempt identifying the foreign key relationship by applying the conditions above. <|eot_id|><|start_header_id|>assistant<|end_header_id|>"""

    LLM_CONFIG = {"temperature": 0.2}

    PROFILING_COLUMNS_REQUIRED = [
        "datatype",
        "distinct_count",
        "uniqueness",
        "completeness",
        "sample_data",
    ]

    def __init__(
        self,
        profiling_data: pd.DataFrame,
        primary_keys: Optional[list[tuple[str, str]]] = None,
        llm=None,
        *args,
        **kwargs,
    ):
        self.CACHE = {}

        self.llm = (
            ChatModelLLM.get_llm(
                model_name=settings.LLM_PROVIDER,
                llm_config=self.LLM_CONFIG,
            )
            if llm is None
            else llm
        )

        self.parser_linkages = JsonOutputParser(pydantic_object=linkage)

        self.prompt_link = PromptTemplate(
            template=self.PROMPT_TEMPLATE,
            input_variables=["current_attempt", "previous_attempts", "table_info"],
            partial_variables={"format_instructions": self.parser_linkages.get_format_instructions()},
        )

        self.link_identifier_llm = self.prompt_link | self.llm | self.parser_linkages

        self.profiling_data = preprocess_profiling_df(profiling_data)
        self.column_datatypes = read_column_datatypes(
            dtype=self.profiling_data[["table_name", "column_name", "datatype_l1"]]
        )

        self.primary_keys = primary_keys if primary_keys is not None else []

        self.table_ddl_statements = {
            table_name: generate_create_table_query(
                gbydata["column_name"].values.tolist(),
                table_name=table_name,
                profiling_data=gbydata,
                primary_keys=self.primary_keys,
                column_datatypes=self.column_datatypes,
                columns_required=self.PROFILING_COLUMNS_REQUIRED,
                mapping_dtypes_to_sql=True,
            )
            for (table_name,), gbydata in self.profiling_data.groupby(["table_name"])
        }

        self.graph = self.__workflow_builder__()

        self.status = None

        self.logs = []

    def link_identifier(self, state):
        """
        Identifies the link between tables
        Args:
            state (dict): The current graph state
        Returns:
            state (dict): New key added to state, Potential_Foreign_Key, that contains the link between two tables
        """

        log.info("---IDENTIFY LINK---")

        input_text = state["input_text"]
        iteration = state["iteration"]

        log_msg = "[*] TRYING TO IDENTIFY LINK BETWEEN: " + str(input_text)
        self.logs.append(log_msg)
        log.info(log_msg)

        if iteration is None:
            iteration = 0

        log_msg = f"[*] Iteration running: {iteration}"
        self.logs.append(log_msg)
        log.info(log_msg)

        if iteration == 0:
            error_msg = [""]
        else:
            error_msg = state["error_msg"]

        tab1, tab2 = input_text.split(" & ")
        table_details = self.table_ddl_statements[tab1] + "\n\n" + self.table_ddl_statements[tab2]

        if iteration == 0:
            current_attempt = 1

        else:
            current_attempt = iteration + 1

        previous_attempt_details = " ".join(error_msg)

        log_msg = f"[*] Message in previous attempt: {previous_attempt_details}"
        self.logs.append(log_msg)
        log.info(log_msg)

        try:
            potential_link = self.link_identifier_llm.invoke({
                "current_attempt": "Attempt " + str(current_attempt),
                "table_info": table_details,
                "previous_attempts": previous_attempt_details,
            })
        except Exception:
            log_msg = "[!] Killing LLM run due to halucinations in output"
            self.logs.append(log_msg)
            self.status = Status.HALLUCINATED_INVOKE
            log.info(log_msg)

            # log.info('[!] Killing LLM run due to halucinations in output')
            potential_link = {
                "table1": "NA",
                "column1": "NA",
                "table2": "NA",
                "column2": "NA",
            }
        iteration = current_attempt

        # Recursively search for the innermost dictionary
        potential_link_innermost_dict = extract_innermost_dict(potential_link)

        # checking for a possibility of recurcive loop in llm
        if iteration > 1:
            previous_potential_link = state["potential_link"]
            if previous_potential_link == potential_link_innermost_dict:
                log_msg = f"[!] Killing LLM run due to possibility of recurssion: {potential_link_innermost_dict}"
                self.status = Status.HALLUCINATED_RECURSION
                log.info(log_msg)
                self.logs.append(log_msg)
                potential_link_innermost_dict = {
                    "table1": "NA",
                    "column1": "NA",
                    "table2": "NA",
                    "column2": "NA",
                }
                return {
                    "potential_link": potential_link_innermost_dict,
                    "iteration": iteration,
                }

        log_msg = "[*] This is the link:" + str(potential_link_innermost_dict)
        log.info(log_msg)
        self.logs.append(log_msg)

        return {"potential_link": potential_link_innermost_dict, "iteration": iteration}

    def single_link_table_column_name_checker(self, state):
        """
        At least one of the columns in link should be unique

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New keys added to state, error_msg & if_error, error message is either table or column is missing and i_error would be True
        """
        # log.info("---CHECKING FOR TABLE & COLUMN NAME PRESENCE---")
        log_msg = "---CHECKING FOR TABLE & COLUMN NAME PRESENCE---"
        log.info(log_msg)
        self.logs.append(log_msg)

        potential_link = state["potential_link"]
        final_error_msg_list = state["error_msg"]
        log.info("chk1")
        # check if table names present
        stat_foreign_key = self.profiling_data.loc[self.profiling_data["table_name"] == potential_link["table1"]]

        if stat_foreign_key.shape[0] == 0:
            log.info("chk2")
            missing_table = potential_link["table1"]
            error_msg = f"{missing_table} seems to be missing. Check for any spelling errors in the table name."
            try:
                final_error_msg_list.append(error_msg)
            except Exception:
                final_error_msg_list = []
                final_error_msg_list.append(error_msg)
            # log.info(f"[!] {error_msg}")
            log_msg = f"[!] {error_msg}"
            log.info(log_msg)
            self.logs.append(log_msg)

            return {"error_msg": final_error_msg_list, "if_error": True}

        log.info("chk3")

        stat_foreign_key = stat_foreign_key.loc[stat_foreign_key["column_name"] == potential_link["column1"],]

        if stat_foreign_key.shape[0] == 0:
            log.info("chk4")
            table_name = potential_link["table1"]
            missing_column = potential_link["column1"]
            error_msg = (
                f"{missing_column} is not present in {table_name}. Check for any spelling errors in the column name."
            )
            try:
                final_error_msg_list.append(error_msg)
            except Exception:
                final_error_msg_list = []
                final_error_msg_list.append(error_msg)
            # log.info(f"[!] {error_msg}")
            log_msg = f"[!] {error_msg}"
            log.info(log_msg)
            self.logs.append(log_msg)

            return {"error_msg": final_error_msg_list, "if_error": True}

        # Referrenced table #########################3
        log.info("chk5")
        # check if table names present
        stat_referrence_key = self.profiling_data.loc[self.profiling_data["table_name"] == potential_link["table2"]]

        if stat_referrence_key.shape[0] == 0:
            log.info("chk6")
            missing_table = potential_link["table2"]
            error_msg = f"{missing_table} seems to be missing. Check for any spelling errors in the table name."
            try:
                final_error_msg_list.append(error_msg)
            except Exception:
                final_error_msg_list = []
                final_error_msg_list.append(error_msg)
            # log.info(error_msg)
            log_msg = f"[!] {error_msg}"
            log.info(log_msg)
            self.logs.append(log_msg)

            return {"error_msg": final_error_msg_list, "if_error": True}
        log.info("chk7")
        # check if column names present
        stat_referrence_key = stat_referrence_key.loc[stat_referrence_key["column_name"] == potential_link["column2"]]
        if stat_referrence_key.shape[0] == 0:
            log.info("chk8")
            table_name = potential_link["table2"]
            missing_column = potential_link["column2"]
            error_msg = (
                f"{missing_column} is not present in {table_name}. Check for any spelling errors in the column name."
            )
            try:
                final_error_msg_list.append(error_msg)
            except Exception:
                final_error_msg_list = []
                final_error_msg_list.append(error_msg)
            # log.info(error_msg)
            log_msg = f"[!] {error_msg}"
            log.info(log_msg)
            self.logs.append(log_msg)

            return {"error_msg": final_error_msg_list, "if_error": True}

        # If no error found
        log_msg = "[*] All Good No error found"
        self.logs.append(log_msg)
        log.info(log_msg)

        # log.info('All Good No error found')
        return {"if_error": False}

    def single_link_uniqueness_checker(self, state):
        """
        At least one of the columns in link should be unique

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New keys added to state, error_msg, error message if one of the column in the link is unique
        """
        # log.info("---CHECKING FOR COLUMN UNIQUENESS---")
        log_msg = "---CHECKING FOR COLUMN UNIQUENESS---"
        log.info(log_msg)
        self.logs.append(log_msg)

        potential_link = state["potential_link"]
        final_error_msg_list = state["error_msg"]

        # For Foreign & Referrenced key table ###########################
        stat_foreign_key = self.profiling_data.loc[
            (self.profiling_data["table_name"] == potential_link["table1"])
            & (self.profiling_data["column_name"] == potential_link["column1"]),
            ["uniqueness_ratio"],
        ]

        stat_referrence_key = self.profiling_data.loc[
            (self.profiling_data["table_name"] == potential_link["table2"])
            & (self.profiling_data["column_name"] == potential_link["column2"]),
            ["uniqueness_ratio"],
        ]

        # Calculate the max uniqueness for the foreign key table
        max_foreign_key_uniqueness = stat_foreign_key["uniqueness_ratio"].max()

        # Calculate the max uniqueness for the referenced key table
        max_referrenced_key_uniqueness = stat_referrence_key["uniqueness_ratio"].max()

        # Optionally, you can calculate the overall max uniqueness
        max_uniqueness = max(max_foreign_key_uniqueness, max_referrenced_key_uniqueness)

        if max_uniqueness >= settings.UNIQUENESS_THRESHOLD:
            # log.info('All Good No error found')
            log_msg = "All Good No error found"
            log.info(log_msg)
            self.logs.append(log_msg)

            return {"if_error": False}
        else:
            ref_column_uniqueness = max_referrenced_key_uniqueness
            ref_table_name = potential_link["table1"]
            ref_column = potential_link["column1"]
            foreign_table_name = potential_link["table2"]
            foreign_column = potential_link["column2"]
            foreign_column_uniqueness = max_foreign_key_uniqueness

            error_msg = f"Uniqueness of {foreign_column} column in table {foreign_table_name} is {foreign_column_uniqueness * 100:.2f} percent, while the uniqueness of {ref_column} in table {ref_table_name} is {ref_column_uniqueness * 100:.2f} percent. This is lower than the acceptable limit"
            try:
                final_error_msg_list.append(error_msg)
            except Exception:
                final_error_msg_list = []
                final_error_msg_list.append(error_msg)

            log_msg = f"[!] {error_msg}"
            log.info(log_msg)
            self.logs.append(log_msg)

            return {"error_msg": final_error_msg_list, "if_error": True}

    def single_link_datatype_checker(self, state):
        """
        Datatype of both columns should be same

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New keys added to state, error_msg, error message datatype of column are not similar
        """
        # log.info("---CHECKING FOR COLUMN UNIQUENESS---")
        log_msg = "---CHECKING FOR COLUMN UNIQUENESS---"
        log.info(log_msg)
        self.logs.append(log_msg)

        potential_link = state["potential_link"]
        final_error_msg_list = state["error_msg"]

        # For Foreign & Referrenced key table ###########################
        stat_foreign_key = self.profiling_data.loc[
            (self.profiling_data["table_name"] == potential_link["table1"])
            & (self.profiling_data["column_name"] == potential_link["column1"]),
            ["datatype_l1"],
        ]

        stat_referrence_key = self.profiling_data.loc[
            (self.profiling_data["table_name"] == potential_link["table2"])
            & (self.profiling_data["column_name"] == potential_link["column2"]),
            ["datatype_l1"],
        ]

        # Calculate the max uniqueness for the foreign key table

        foreign_key_datatype = stat_foreign_key["datatype_l1"].values[0]

        # Calculate the max uniqueness for the referenced key table

        referrenced_key_datatype = stat_referrence_key["datatype_l1"].values[0]

        if dtype_check(dtype1=foreign_key_datatype, dtype2=referrenced_key_datatype):
            # log.info('All Good No error found')
            log_msg = "All Good No error found"
            log.info(log_msg)
            self.logs.append(log_msg)

            return {"if_error": False}
        else:
            ref_table_name = potential_link["table1"]
            ref_column = potential_link["column1"]
            foreign_table_name = potential_link["table2"]
            foreign_column = potential_link["column2"]

            error_msg = f"Datatype of {foreign_column} column in table {foreign_table_name} is {foreign_key_datatype}, while the datatype of {ref_column} in table {ref_table_name} is {referrenced_key_datatype}. To join two columns dataype of both should be same"
            try:
                final_error_msg_list.append(error_msg)
            except Exception:
                final_error_msg_list = []
                final_error_msg_list.append(error_msg)

            log.info(error_msg)
            self.logs.append(f"[!] {error_msg}")

            return {"error_msg": final_error_msg_list, "if_error": True}

    def single_link_intersection_checker(self, state):
        """
        Checks intersection of values between columns

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New keys added to state, error_msg, error message if intersection between both the columns is low
        """
        log_msg = "---CHECKING FOR INTERSECTION BETWEEN COLUMNS---"
        log.info(log_msg)
        self.logs.append(log_msg)

        potential_link = state["potential_link"]
        final_error_msg_list = state["error_msg"]

        table1_name = potential_link["table1"]
        column1_name = potential_link["column1"]
        table2_name = potential_link["table2"]
        column2_name = potential_link["column2"]

        count_distinct_col1 = self.profiling_data.loc[
            (self.profiling_data.table_name == table1_name)
            & (self.profiling_data.column_name == column1_name)
        ]["distinct_count"].values[0]
        count_distinct_col2 = self.profiling_data.loc[
            (self.profiling_data.table_name == table2_name)
            & (self.profiling_data.column_name == column2_name)
        ]["distinct_count"].values[0]

        # Use adapter from one of the datasets to execute query
        adapter = self.table1.adapter

        try:
            intersect_count = adapter.intersect_count(
                table1=self.table1,
                column1_name=column1_name,
                table2=self.table2,
                column2_name=column2_name
            )
        except Exception as e:
            log.error(f"Error executing intersection query: {e}")
            error_msg = f"Could not calculate intersection between {table1_name}.{column1_name} and {table2_name}.{column2_name}."
            try:
                final_error_msg_list.append(error_msg)
            except AttributeError:
                final_error_msg_list = [error_msg]
            return {"error_msg": final_error_msg_list, "if_error": True}

        intersect_ratio_from_col = intersect_count / count_distinct_col1 if count_distinct_col1 > 0 else 0
        intersect_ratio_to_col = intersect_count / count_distinct_col2 if count_distinct_col2 > 0 else 0

        if intersect_count == 0:
            error_msg = f"The intersection between {column1_name} column in table {table1_name} with {column2_name} in table {table2_name} resulted in zero rows"
            try:
                final_error_msg_list.append(error_msg)
            except AttributeError:
                final_error_msg_list = [error_msg]
            log.info(error_msg)
            self.logs.append(error_msg)
            return {"error_msg": final_error_msg_list, "if_error": True}

        elif max(intersect_ratio_from_col, intersect_ratio_to_col) < settings.INTERSECT_RATIO_THRESHOLD:
            error_msg1 = f"Only {intersect_ratio_from_col * 100:.2f} percent of values in {column1_name} in table {table1_name} matched with {column2_name} in table {table2_name}. "
            error_msg2 = f"Only {intersect_ratio_to_col * 100:.2f} percent of values in {column2_name} in table {table2_name} matched with {column1_name} in table {table1_name}."
            error_msg = error_msg1 + error_msg2
            try:
                final_error_msg_list.append(error_msg)
            except AttributeError:
                final_error_msg_list = [error_msg]
            log.info(error_msg)
            self.logs.append(error_msg)
            return {"error_msg": final_error_msg_list, "if_error": True}
        else:
            log_msg = "All Good No error found"
            log.info(log_msg)
            self.logs.append(log_msg)
            return {
                "if_error": False,
                "intersect_count": intersect_count,
                "intersect_ratio_from_col": round(intersect_ratio_from_col, 3),
                "intersect_ratio_to_col": round(intersect_ratio_to_col, 3),
                "accuracy": round(max(intersect_ratio_from_col, intersect_ratio_to_col), 3),
            }

    def link_check_router(self, state):
        """
        Route the link checks

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New key added to state, link_check_type, that contains the type of link which we need to check
        """
        # log.info("---IDENTIFY TYPE OF LINK---")
        log_msg = "---IDENTIFY TYPE OF LINK---"
        log.info(log_msg)
        self.logs.append(log_msg)

        potential_link = state["potential_link"]

        if potential_link["table1"] == "NA":
            link_type = "No_Link"
            if self.status is None:
                self.status = Status.NO_LINKS

        elif (len(potential_link) > 0) & (len(potential_link["table1"]) > 1):
            link_type = "Single_Link"
            self.status = Status.COMPLETED
        else:
            link_type = "Multiple_Link"
            self.status = Status.COMPLETED

        log_msg = f"Link type: {link_type}"
        log.info(log_msg)
        self.logs.append(log_msg)

        return link_type

    def error_checker(self, state):
        """
        Route to link identifier if error found

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New key added to state, if_error, binary response for error present
        """
        log_msg = "---DID WE FIND ANY ERROR?---"
        log.info(log_msg)
        self.logs.append(log_msg)

        if_error = state["if_error"]
        if not if_error:
            error_found = "No"
        else:
            error_found = "Yes"
        log.info(error_found)
        return error_found

    def __workflow_builder__(self):
        log.info("[*] Compiling Agent Workflow Graph")

        from langgraph.graph import END, START, StateGraph

        from .utils import GraphState

        # Node adding
        workflow = StateGraph(GraphState)
        workflow.add_node("link_identifier", self.link_identifier)  # link identifier
        workflow.add_node(
            "single_link_table_column_name_checker",
            self.single_link_table_column_name_checker,
        )  # table column name checker
        workflow.add_node("single_link_uniqueness_checker", self.single_link_uniqueness_checker)  # uniqueness
        workflow.add_node("single_link_datatype_checker", self.single_link_datatype_checker)  # datatype
        workflow.add_node("single_link_intersection_checker", self.single_link_intersection_checker)

        # Adding edges
        workflow.add_edge(START, "link_identifier")

        workflow.add_conditional_edges(
            "link_identifier",
            self.link_check_router,
            {
                "Single_Link": "single_link_table_column_name_checker",
                "Multiple_Link": END,
                "No_Link": END,
            },
        )

        workflow.add_conditional_edges(
            "single_link_table_column_name_checker",
            self.error_checker,
            {
                "Yes": "link_identifier",
                "No": "single_link_uniqueness_checker",
            },
        )

        workflow.add_conditional_edges(
            "single_link_uniqueness_checker",
            self.error_checker,
            {
                "Yes": "link_identifier",
                "No": "single_link_datatype_checker",
            },
        )

        workflow.add_conditional_edges(
            "single_link_datatype_checker",
            self.error_checker,
            {
                "Yes": "link_identifier",
                "No": "single_link_intersection_checker",
            },
        )

        workflow.add_conditional_edges(
            "single_link_intersection_checker",
            self.error_checker,
            {
                "Yes": "link_identifier",
                "No": END,
            },
        )

        # Compile the workflow into a LangChain Runnable
        graph = workflow.compile()

        return graph

    def __post_processing__(self, data: pd.DataFrame) -> pd.DataFrame:
        def process(result: str):
            try:
                result = eval(result) if isinstance(result, str) else result
                return result if result["table1"] != "NA" else ""
            except Exception as ex:
                log.info(f"[!] Error while parsing: {ex}")
                return ""

        data["links"] = data.links.apply(process)
        res = data.links.tolist()
        res = list(filter(lambda x: not isinstance(x, str), res))
        res = pd.DataFrame(res)
        if res.shape[0] != 0:
            res.rename(
                columns={
                    "table1": "table1_name",
                    "column1": "column1_name",
                    "table2": "table2_name",
                    "column2": "column2_name",
                },
                inplace=True,
            )
        return res

    def __graph_invoke__(self, table1: DataSet, table2: DataSet) -> dict:
        self.table1 = table1
        self.table2 = table2
        final_output = {}
        final_output["table1"] = table1.name
        final_output["table2"] = table2.name

        start_time = time.time()
        input_message = HumanMessage(content=f"{table1.name} & {table2.name}")
        init_data = {
            "input_text": input_message.content,
            "iteration": None,
            "error_msg": [""],
        }

        try:
            event = {}
            for event in self.graph.stream(
                init_data,
                stream_mode="values",
                config={
                    "recursion_limit": 20,
                    "metadata": {"table_combo": tuple(sorted([table1.name, table2.name]))},
                },
            ):
                for key in [list(event.keys())[-1]]:
                    log.info(f"Finished running: {key}:")

        except GraphRecursionError as ex:
            log.warning(f"[!] Graph went into recursion loop when running for {table1.name} <=> {table2.name}")
            event["status"] = Status.HALLUCINATED_RECURSION
            event["potential_link"] = "NA"
            event["error_msg"] = ex

        except Exception as ex:
            import traceback
            log.error(f"[!] Error while running for {table1.name} <=> {table2.name}: Reason {traceback.format_exc()}")
            event["status"] = Status.HALLUCINATED_RECURSION
            event["potential_link"] = "NA"
            event["error_msg"] = ex
        # End time
        end_time = time.time()

        # Compute runtime
        runtime = end_time - start_time
        log.info(f"Runtime: {runtime:.2f} seconds")
        log.info(f"[*] link between tables: {event['potential_link']}")
        potential_link = event["potential_link"]

        if potential_link and potential_link.get("table1") != "NA":
            potential_link["intersect_count"] = event.get("intersect_count")
            potential_link["intersect_ratio_from_col"] = event.get("intersect_ratio_from_col")
            potential_link["intersect_ratio_to_col"] = event.get("intersect_ratio_to_col")
            potential_link["accuracy"] = event.get("accuracy")

        final_output["links"] = potential_link
        final_output["Runtime_secs"] = runtime
        # Logs monitoring
        final_output["logs"] = "\n".join(self.logs)
        final_output["status"] = self.status
        final_output["validation_logs"] = event["error_msg"]

        return final_output

    def __call__(self, assets: List[Tuple[DataSet]], *args, **kwds):
        results = []
        pb = tqdm(assets)
        for _, (table1, table2) in enumerate(pb):
            runs = 1
            pb.set_description(desc=f"[*] {table1.name} <==> {table2.name} Runs:{runs}")

            final_output = self.__graph_invoke__(table1, table2)

            # If status came out as Hallucinated retry one more time
            while (
                final_output["status"] in (Status.HALLUCINATED_INVOKE, Status.HALLUCINATED_RECURSION)
                and runs < self.HALLUCINATIONS_MAX_RETRY
            ):
                runs += 1
                log.info(f"[*] Hallucinated for {table1.name} <==> {table2.name} ... Retry no {runs} ")
                final_output = self.__graph_invoke__(table1=table1, table2=table2)

            results.append(final_output)
        return self.__post_processing__(data=pd.DataFrame(results)) if len(
            results
        ) != 0 else pd.DataFrame(), pd.DataFrame(results)



from intugle.common.exception import errors
from intugle.models.manifest import Manifest


class TableSchema:
    """Class to generate and manage SQL table schemas based on a manifest."""

    def __init__(self, manifest: Manifest):
        """
        Initializes the TableSchema with a manifest.
        
        Args:
            manifest (Manifest): The manifest containing the details of the tables.
        """
        self.manifest = manifest
        self.table_schemas: dict[str, str] = {}

    def generate_table_schema(self, table_name: str) -> str:
        """Generate the SQL schema for a given table based on its details in the manifest.

        Args:
            table_name (str): The name of the table for which to generate the schema.

        Returns:
            str: The SQL schema definition for the table.
        """
        table_detail = self.manifest.sources.get(table_name)
        if not table_detail:
            raise errors.NotFoundError(f"Table {table_name} not found in manifest.")

        # Start with the CREATE TABLE statement
        schema = f"CREATE TABLE {table_detail.table.name} -- {table_detail.table.description}"

        # Iterate through the columns of the table and create the column definitions
        columns_statements = [
            f"\"{column.name}\" {column.type}, -- {column.description}" for column in table_detail.table.columns
        ]

        # join the column definitions into a single string
        column_schema = "\n".join(columns_statements)

        # Add the column definitions to the schema
        schema += "\n(" + column_schema + "\n);"

        return schema

    def get_table_schema(self, table_name: str):
        """Get the SQL schema for a specified table, generating it if not already cached.

        Args:
            table_name (str): The name of the table for which to retrieve the schema.

        Returns:
            str: The SQL schema definition for the table.
        """
        table_schema = self.table_schemas.get(table_name)

        if table_schema is None:
            table_schema = self.generate_table_schema(table_name)
            self.table_schemas[table_name] = table_schema

        return table_schema



# ====================================================================
#  Importing the required python packages
# ====================================================================

import logging
import multiprocessing
import re
import string

import numpy as np
import pandas as pd

# import pyarrow
# ====================================================================
#  Importing the required module packages
# ====================================================================
# from legoai.datatype_identification.bag_of_characters import extract_bag_of_characters_features
# from legoai.datatype_identification.bag_of_words import extract_bag_of_words_features
# from legoai.datatype_identification.word_embeddings import extract_word_embeddings_features
# from legoai.datatype_identification.custom_features import extract_addl_feats
# from legoai.datatype_identification.paragraph_vectors import infer_paragraph_embeddings_features
# from legoai.datatype_identification.global_state import set_first, reset_first
# from legoai.datatype_identification.helpers import literal_eval_as_str, keys_to_csv
# from pyarrow.parquet import ParquetFile
from intugle.core.settings import settings

log = logging.getLogger(__name__)


# Ignore list for the preprocessing values
ignoreList = ['#na', '#n/a', 'na', 'n/a', 'none', 'nan', 'blank', 'blanks', 'nil', 'n.a.', 'n.a',
              '"#na"', '"#n/a"', '"na"', '"n/a"', '"none"', '"nan"', '"blank"', '"blanks"', '"nil"', '"n.a."', '"n.a"',
              "'#na'", "'#n/a'", "'na'", "'n/a'", "'none'", "'nan'", "'blank'", "'blanks'", "'nil'", "'n.a.'", "'n.a'"]

core_count = multiprocessing.cpu_count()  
size = settings.DI_CONFIG['THRESHOLD']['PARTITION_SIZE']


# ====================================================================
# normalise_whitespace: 
#   - Clean whitespace from strings by:
#   - trimming leading and trailing whitespace
#   - normalising all whitespace to spaces
#   - reducing whitespace sequences to a single space
# Parameters: 
#     data - input data
# ====================================================================
def normalise_whitespace(data) -> str:
    if isinstance(data, str):
        return re.sub(r"\s{2,}", " ", data.strip())
    else:
        return data


def normalise_string_whitespace(col_values) -> list:
    
    # Get the metadata info such as id, dataset, table, column name
    table_name = col_values[0]
    column_name = col_values[1]
    if len(col_values[2:]) > 0: 
        normalized_values = list(np.vectorize(normalise_whitespace)(np.array(col_values[5:])))
        # normalized_values = list(pseq(map(additional_processing, col_values[5:]), processes=core_count, partition_size=size))
        # Remove the whitespaces from the data
        # normalized_values = list(map(normalise_whitespace, col_values[5:]))

        # Removing the table and column name from values ## Added to remove features list
        normalized_values = np.vectorize(str)(np.array(normalized_values))

        # Convert to lowercase
        lowercase_values = np.char.lower(normalized_values)

        # # Filter out the unwanted values
        mask = ~np.isin(lowercase_values, [table_name.lower(), column_name.lower()])

        # # Apply the mask
        normalized_values = list(normalized_values[mask])

        # normalized_values = [val for val in normalized_values if
        #                      str(val).lower() not in [dataset_name.lower(), table_name.lower(), column_name.lower()]]

        # Combining the metadata with the normalized values into a list of data
        
    else:
        log.warning(f"[!] Empty column encountered for {table_name} ==> {column_name} ...")
        normalized_values = ['']
        
    normalized_values_upd = [table_name] + [column_name] + normalized_values
    return normalized_values_upd


# ====================================================================
# remove_table_column_name: 
#     -  Remove punctuation characters at start and end of the data
# Parameters: 
#     strs - input string
# ====================================================================
def remove_table_column_name(values: str, table_name: str, column_name: str):
    return [val for val in values if
            str(val).lower() not in [table_name.lower(), column_name.lower()]]


# ====================================================================
# removeASCII: 
#     -  Remove ASCII Characters from the data
# Parameters: 
#     strs - string data for cleaning ascii characters
# ====================================================================
def removeASCII(strs) -> str:
    return ''.join([char for word in str(strs) for char in word if ord(char) < 128])


# ====================================================================
# removePunctuation: 
#     -  Remove punctuation characters at start and end of the data
# Parameters: 
#     strs - input string
# ====================================================================
def removePunctuation(strs) -> str:
    return strs.strip("'").strip('"')


# ====================================================================
# removePunctuation: 
#     - Remove punctuation from the text and check if they are empty
#     - If it contains only punctuations, then return empty else return string 
# Parameters: 
#     strs - input string
# ====================================================================
def removePunctuationText(strs) -> str:
    clean_str = strs.translate(str.maketrans('', '', string.punctuation))
    clean_str = normalise_whitespace(clean_str)
    clean_str = clean_str.strip()
    return '' if len(clean_str) == 0 else strs  # Checks if the string is empty else return string


# ====================================================================
# additional_processing: 
#     - Remove null, nan, none and ignore list elements from data
#     - Remove ASCII Characters and unicode characters
#     - Remove punctuation text and remove punctuation in the start and end text
# Parameters: 
#     value - list of data
# ====================================================================
def additional_processing(value) -> str:

    # Remove the nulls/none and not in ignore list
    if value is None or pd.isnull(value) or str(value).lower() in ignoreList:
        return_val = ''
    else:
        value = str(value).replace('\xa0', ' ').strip()
        return_val = removeASCII(value)

    # Remove punctuation in the start and end text
    return_val = removePunctuationText(return_val)
    
    # Remove punctuation text from text
    return_val = removePunctuation(return_val)
    return return_val


# ====================================================================
# special_token_repl: 
#     -  Remove punctuation characters at start and end of the data
# Parameters: 
#     text - input string
#     suffix - additional text to be added at the end
# ====================================================================
def special_token_repl(text: str, suffix: str) -> str:
    
    # Replace the pattern with empty space and replace multiple space with single space
    replaced_text = text.translate(str.maketrans(string.punctuation, ' ' * len(string.punctuation)))
    replaced_text = re.sub(string=replaced_text, pattern=' +', repl=' ')

    # Check if the replaced text is empty then add unknown and suffix word
    if replaced_text == '':
        replaced_text = 'unknown' + suffix

    return replaced_text



from enum import Enum
from typing import Optional

from intugle.common.resources.base import BaseResource
from intugle.common.schema import NodeType, SchemaBase
from intugle.libs.smart_query_generator.models.models import LinkModel


class RelationshipTable(SchemaBase):
    table: str
    column: str


class RelationshipProfilingMetrics(SchemaBase):
    intersect_count: Optional[int] = None
    intersect_ratio_from_col: Optional[float] = None
    intersect_ratio_to_col: Optional[float] = None
    accuracy: Optional[float] = None


class RelationshipType(str, Enum):
    ONE_TO_ONE = "one_to_one"
    ONE_TO_MANY = "one_to_many"
    MANY_TO_ONE = "many_to_one"
    MANY_TO_MANY = "many_to_many"


class Relationship(BaseResource):
    resource_type: NodeType = NodeType.RELATIONSHIP
    source: RelationshipTable
    target: RelationshipTable
    profiling_metrics: Optional[RelationshipProfilingMetrics] = None
    type: RelationshipType

    @property
    def link(self) -> LinkModel:
        source_field_id = f"{self.source.table}.{self.source.column}"
        target_field_id = f"{self.target.table}.{self.target.column}"
        link: LinkModel = LinkModel(
            id=self.name,
            source_field_id=source_field_id,
            source_asset_id=self.source.table,
            target_field_id=target_field_id,
            target_asset_id=self.target.table,
        )
        return link



import numpy as np
import pandas as pd

from intugle.core.semantic_search.models import RelevancyCategory


def relevancy_adder(
    result: pd.DataFrame,
    relevancy_scores: dict,
    o_column="relevancy",
    i_column="score",
) -> pd.DataFrame:

    # Add relevancy to each terms identified
    result[o_column] = np.where(
        result[i_column] >= relevancy_scores[RelevancyCategory.MOST_RELEVANT],
        RelevancyCategory.MOST_RELEVANT,
        np.where(
            np.logical_and(
                result[i_column] >= relevancy_scores[RelevancyCategory.RELEVANT],
                result[i_column] < relevancy_scores[RelevancyCategory.MOST_RELEVANT],
            ),
            RelevancyCategory.RELEVANT,
            np.where(
                np.logical_and(
                    result[i_column] >= relevancy_scores[RelevancyCategory.LESS_RELEVANT],
                    result[i_column] < relevancy_scores[RelevancyCategory.RELEVANT],
                ),
                RelevancyCategory.LESS_RELEVANT,
                RelevancyCategory.NON_RELEVANT,
            ),
        ),
    )
    return result


def batched(data: pd.DataFrame, n):
    for index in range(0, data.shape[0], n):
        yield data[index : index + n]




from typing import Any

import numpy as np


def convert_to_native(value: Any) -> Any:
    """Recursively converts numpy types to native Python types."""
    if isinstance(value, np.generic):
        return value.item()
    if isinstance(value, (list, tuple)):
        return [convert_to_native(v) for v in value]
    return value


import logging

import pandas as pd

from intugle.core.settings import settings

from .feature_creation import DIFeatureCreation
from .initialize import di_initalizer
from .l1_model import L1Model

log = logging.getLogger(__name__)


MODEL_VERSION = "13052023"


class DataTypeIdentificationPipeline:

    # NUM_SAMPLES = DI_CONFIG["L2PARAMS"]["SAMPLE_SIZE"]

    VALID_INVALID_MAPPING = {
        "integer": ["intRatio"],
        "others": [],
        "float": ["floatRatio"],
        "range_type": ["rangeRatio"],
        "date & time": ["dateRatio"],
        "close_ended_text": ["alphaRatio", "numericRatio", "alphaNumRatio"],
        "open_ended_text": ["alphaRatio", "numericRatio", "alphaNumRatio"],
        "alphanumeric": ["alphaNumRatio", "alphaRatio", "numericRatio"],
    }

    def __init__(self, l1_model: L1Model = None, feature_creator: DIFeatureCreation = None, *args, **kwargs):
        
        di_initalizer()

        self.__feature_creator = DIFeatureCreation() if feature_creator is None else feature_creator
        
        self.__l1_model = L1Model() if l1_model is None else l1_model

    def create_feature(self, assets: list[dict],) -> pd.DataFrame:
        """
        Run DI Feature creation process for a dataset
        Parameters
        ----------
        source_path (os.path): full folder path to the dataset.
        parallel (bool): flag for whether to run parallel processing on feature creation or not.
        batched (bool): flag to whether run features creation on batches of columns.

        Returns
        -------
        pd.DataFrame: DI features as a pandas dataframe where each row is a feature of a specific column , of a specific table.
        """
        return self.__feature_creator(assets=assets,)

    def l1_model_run(self, features: pd.DataFrame, *args, **kwargs) -> pd.DataFrame:
        """
        Runs DI L1 model to identify datatypes

        Parameters
        ----------
        features (pd.DataFrame): DI specific features obtained from create_feature()
        model_version (str): version of model to run L1 Classifier for.

        Returns
        -------
        pd.DataFrame: predicted datatypes L1 as a dataframe
        """
        return self.__l1_model.model_prediction(features, model_version=kwargs.get("model_version", MODEL_VERSION),)
    
    def __call__(self, sample_values_df: pd.DataFrame, *args, **kwargs):

        feature = self.__feature_creator(sample_values_df=sample_values_df)
        
        log.info("[*] Running L1 Model")
        
        l1_pred_result = self.l1_model_run(feature, model_version=settings.DI_MODEL_VERSION, *args, **kwargs)

        log.info("[*] Merging L1 result and L1 features")
        
        l1_pred_result = pd.merge(l1_pred_result, feature, on=['table_name', 'column_name'])
        
        del feature
 
        return l1_pred_result
    


from typing import Iterable, Iterator, List, MutableSet, Optional, Sequence, TypeVar, overload

SLICE_ALL = slice(None)
_T = TypeVar("_T")


class OrderedSet(MutableSet[_T], Sequence[_T]):
    def __init__(self, initalize: Optional[List[_T]] = None):
        if initalize is None:
            initalize = []
        self.items = initalize

    def add(self, element: _T) -> None:
        if element not in self.items:
            self.items.append(element)

    append = add

    def remove(self, element: _T) -> None:
        if element in self.items:
            self.items.remove(element)

    discard = remove

    def __iter__(self) -> Iterator[_T]:
        return iter(self.items)

    def __len__(self) -> int:
        """
        Returns the number of unique elements in the ordered set

        Example:
            >>> len(OrderedSet([]))
            0
            >>> len(OrderedSet([1, 2]))
            2
        """
        return len(self.items)

    @overload
    def __getitem__(self, index: slice) -> "OrderedSet[_T]": ...

    @overload
    def __getitem__(self, index: Sequence[int]) -> List[_T]: ...

    @overload
    def __getitem__(self, index: int) -> _T: ...

    # concrete implementation
    def __getitem__(self, index):
        """
        Get the item at a given index.

        If `index` is a slice, you will get back that slice of items, as a
        new OrderedSet.

        If `index` is a list or a similar iterable, you'll get a list of
        items corresponding to those indices. This is similar to NumPy's
        "fancy indexing". The result is not an OrderedSet because you may ask
        for duplicate indices, and the number of elements returned should be
        the number of elements asked for.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset[1]
            2
        """
        if isinstance(index, slice) and index == SLICE_ALL:
            return self.copy()
        elif isinstance(index, Iterable):
            return [self.items[i] for i in index]
        elif isinstance(index, slice) or hasattr(index, "__index__"):
            result = self.items[index]
            if isinstance(result, list):
                return self.__class__(result)
            else:
                return result
        else:
            raise TypeError("Don't know how to index an OrderedSet by %r" % index)

    def __repr__(self) -> str:
        if not self:
            return f"{self.__class__.__name__}()"
        return f"{self.__class__.__name__}({list(self)!r})"

    def copy(self) -> "OrderedSet[_T]":
        """
        Return a shallow copy of this object.

        Example:
            >>> this = OrderedSet([1, 2, 3])
            >>> other = this.copy()
            >>> this == other
            True
            >>> this is other
            False
        """
        return self.__class__(self)



import logging
import yaml

from typing import TYPE_CHECKING, Any, Dict, List

import pandas as pd

from intugle.analysis.models import DataSet
from intugle.core.console import console, success_style
from intugle.link_predictor.predictor import LinkPredictor
from intugle.semantic_search import SemanticSearch
from intugle.exporters.factory import factory as exporter_factory

if TYPE_CHECKING:
    from intugle.link_predictor.models import PredictedLink
    from intugle.adapters.adapter import Adapter

log = logging.getLogger(__name__)


class SemanticModel:
    def __init__(self, data_input: Dict[str, Any] | List[DataSet], domain: str = ""):
        self.datasets: Dict[str, DataSet] = {}
        self.links: list[PredictedLink] = []
        self.domain = domain
        self._semantic_search_initialized = False

        if isinstance(data_input, dict):
            self._initialize_from_dict(data_input)
        elif isinstance(data_input, list):
            self._initialize_from_list(data_input)
        else:
            raise TypeError("Input must be a dictionary of named dataframes or a list of DataSet objects.")

    def _initialize_from_dict(self, data_dict: Dict[str, Any]):
        """Creates and processes DataSet objects from a dictionary of raw dataframes."""
        for name, df in data_dict.items():
            dataset = DataSet(df, name=name)
            self.datasets[name] = dataset

    def _initialize_from_list(self, data_list: List[DataSet]):
        """Processes a list of existing DataSet objects"""
        for dataset in data_list:
            if not dataset.name:
                raise ValueError("DataSet objects provided in a list must have a 'name' attribute.")
            self.datasets[dataset.name] = dataset

    def profile(self, force_recreate: bool = False):
        """Run profiling, datatype identification, and key identification for all datasets."""
        console.print("Starting profiling and key identification stage...", style="yellow")
        for dataset in self.datasets.values():
            # Check if this stage is already complete
            if dataset.source_table_model.key is not None and not force_recreate:
                print(f"Dataset '{dataset.name}' already profiled. Skipping.")
                continue

            console.print(f"Processing dataset: {dataset.name}", style="orange1")
            dataset.profile(save=True)
            dataset.identify_datatypes(save=True)
            dataset.identify_keys(save=True)
        console.print("Profiling and key identification complete.", style="bold green")

    def predict_links(self, force_recreate: bool = False):
        """Run link prediction across all datasets."""
        console.print("Starting link prediction stage...", style="yellow")
        self.link_predictor = LinkPredictor(list(self.datasets.values()))
        self.link_predictor.predict(save=True, force_recreate=force_recreate)
        self.links: list[PredictedLink] = self.link_predictor.links
        console.print("Link prediction complete.", style="bold green")

    def generate_glossary(self, force_recreate: bool = False):
        """Generate business glossary for all datasets."""
        console.print("Starting business glossary generation stage...", style="yellow")
        for dataset in self.datasets.values():
            # Check if this stage is already complete
            if dataset.source_table_model.description and not force_recreate:
                console.print(f"Glossary for '{dataset.name}' already exists. Skipping.")
                continue

            console.print(f"Generating glossary for dataset: {dataset.name}", style=success_style)
            dataset.generate_glossary(domain=self.domain, save=True)
        console.print("Business glossary generation complete.", style="bold green")

    def build(self, force_recreate: bool = False):
        """Run the full end-to-end knowledge building pipeline."""
        self.profile(force_recreate=force_recreate)
        self.predict_links()
        self.generate_glossary(force_recreate=force_recreate)

        # Initialize semantic search
        try:
            self.initialize_semantic_search()
        except Exception as e:
            log.warning(f"Semantic search initialization failed during build: {e}")

        return self

    def export(self, format: str, **kwargs):
        """Export the semantic model to a specified format."""
        # This assumes that the manifest is already loaded in the SemanticModel
        # In a real implementation, you would get the manifest from the SemanticModel instance
        from intugle.parser.manifest import ManifestLoader
        from intugle.core import settings
        manifest_loader = ManifestLoader(settings.PROJECT_BASE)
        manifest_loader.load()
        manifest = manifest_loader.manifest

        exporter = exporter_factory.get_exporter(format, manifest)
        exported_data = exporter.export(**kwargs)
        
        output_path = kwargs.get("path")
        if output_path:
            with open(output_path, "w") as f:
                yaml.dump(exported_data, f, sort_keys=False, default_flow_style=False)
            print(f"Successfully exported to {output_path}")
        
        return exported_data

    @property
    def profiling_df(self) -> pd.DataFrame:
        """Returns a consolidated DataFrame of profiling metrics for all datasets."""
        all_profiles = [dataset.profiling_df for dataset in self.datasets.values()]
        return pd.concat(all_profiles, ignore_index=True)

    @property
    def links_df(self) -> pd.DataFrame:
        """Returns the predicted links as a pandas DataFrame."""
        if hasattr(self, "link_predictor"):
            return self.link_predictor.get_links_df()
        return pd.DataFrame()

    @property
    def glossary_df(self) -> pd.DataFrame:
        """Returns a consolidated DataFrame of glossary information for all datasets."""
        glossary_data = []
        for dataset in self.datasets.values():
            for column in dataset.source_table_model.columns:
                glossary_data.append(
                    {
                        "table_name": dataset.name,
                        "column_name": column.name,
                        "column_description": column.description,
                        "column_tags": column.tags,
                    }
                )
        return pd.DataFrame(glossary_data)

    def initialize_semantic_search(self):
        """Initialize the semantic search engine."""
        try:
            print("Initializing semantic search...")
            search_client = SemanticSearch()
            search_client.initialize()
            self._semantic_search_initialized = True
            print("Semantic search initialized.")
        except Exception as e:
            log.warning(f"Could not initialize semantic search: {e}")
            raise e

    def visualize(self):
        return self.link_predictor.show_graph()

    def search(self, query: str):
        """Perform a semantic search on the knowledge base."""
        if not self._semantic_search_initialized:
            self.initialize_semantic_search()

        try:
            search_client = SemanticSearch()
            return search_client.search(query)
        except Exception as e:
            log.error(f"Could not perform semantic search: {e}")
            raise e

    def deploy(self, target: str, **kwargs):
        """
        Deploys the semantic model to a specified target platform based on the persisted YAML files.

        Args:
            target (str): The target platform to deploy to (e.g., "snowflake").
            **kwargs: Additional keyword arguments specific to the target platform.
        """
        console.print(f"Starting deployment to '{target}' based on project YAML files...", style="yellow")

        # 1. Load the entire project state from YAML files
        from intugle.parser.manifest import ManifestLoader
        from intugle.core import settings
        manifest_loader = ManifestLoader(settings.PROJECT_BASE)
        manifest_loader.load()
        manifest = manifest_loader.manifest

        # 2. Find a suitable adapter from the loaded manifest
        adapter_to_use: "Adapter" = None
        
        # Dynamically get the adapter class from the factory
        from intugle.adapters.factory import AdapterFactory
        factory = AdapterFactory()
        
        target_adapter_class = None
        for name, (checker, creator) in factory.dataframe_funcs.items():
            if name == target.lower():
                target_adapter_class = creator
                break

        if not target_adapter_class:
            raise ValueError(f"Deployment target '{target}' is not supported or its dependencies are not installed.")

        # Find a source that matches the target type to instantiate the adapter
        for source in manifest.sources.values():
            if source.table.details and source.table.details.get("type") == target.lower():
                adapter_to_use = target_adapter_class()
                break
        
        if not adapter_to_use:
            raise RuntimeError(
                f"Cannot deploy to '{target}'. No '{target}' source found in the project YAML files "
                "to provide connection details."
            )


        # 4. Delegate the deployment to the adapter, passing full manifest
        try:
            adapter_to_use.deploy_semantic_model(manifest, **kwargs)
            console.print(f"Successfully deployed semantic model to '{target}'.", style="bold green")
        except Exception as e:
            console.print(f"Failed to deploy semantic model to '{target}': {e}", style="bold red")
            raise



import asyncio
import logging
import sys

from typing import Optional

import numpy as np
import pandas as pd

from qdrant_client import models

from intugle.core.llms.embeddings import Embeddings, EmbeddingsType
from intugle.core.semantic_search.models import RelevancyCategory
from intugle.core.semantic_search.utils import relevancy_adder
from intugle.core.settings import settings
from intugle.core.vector_store import VectorStoreService
from intugle.core.vector_store.qdrant import VectorSearchKwargs

log = logging.getLogger(__name__)


# class HybridDenseLateSearch:
#     def __init__(self):
#         pass


class HybridDenseLateSearch:
    def __init__(
        self,
        collection_name: str,
        embeddings: Embeddings,
        threshold_for_less_relevant_result: float = 0.5,
        relevancy_config: Optional[dict] = None,
    ):
        self.collection_name = collection_name
        self.embeddings = embeddings
        self.threshold_for_less_relevant_result = threshold_for_less_relevant_result

        if relevancy_config is None:
            relevancy_config = (
                {
                    RelevancyCategory.MOST_RELEVANT: 0.85,
                    RelevancyCategory.RELEVANT: 0.8,
                    RelevancyCategory.LESS_RELEVANT: 0.75,
                }
            )
        self.relevancy_config = relevancy_config

    @property
    def vector_store(self):
        client_config = {"url": settings.QDRANT_URL, "api_key": settings.QDRANT_API_KEY}
        return VectorStoreService(
            collection_name=self.collection_name, collection_configurations=None, client_config=client_config
        )

    async def __vector_search__(
        self,
        query_vector: np.ndarray | list,
        using: str,
        score_threshold: float = None,
        query_filter: models.Filter = None,
        limit: int = sys.maxsize,
    ) -> pd.DataFrame:
        query_vector = query_vector.tolist() if isinstance(query_vector, np.ndarray) else query_vector

        search_result = pd.DataFrame()

        qdrant_points = await self.vector_store.search(
            query=query_vector,
            search_using=using,
            includes=["metadata"],
            search_kwargs=dict(
                VectorSearchKwargs(
                    search_type="similarity",
                    top_k=limit,
                    filter=query_filter,
                    search_params=models.SearchParams(
                        hnsw_ef=limit,
                        exact=True,
                    ),
                    score_threshold=score_threshold,
                ),
            ),
        )

        search_result = pd.DataFrame(qdrant_points.model_dump()["points"])

        if search_result.shape[0] > 0:
            search_result = search_result[["id", "score", "payload"]].reset_index(drop=True)

            search_result = pd.concat([search_result, pd.json_normalize(search_result.payload)], axis=1)[
                ["score", "column_id", "type"]
            ]

        else:
            log.warning(f"[!] No results found when searched {using}")

        return search_result

    async def fetch(
        self,
        # query_term: str,
        dense_vector: np.ndarray,
        late_vector: np.ndarray,
        add_relevancy: bool = True,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        hybrid_result = pd.DataFrame()

        filter_condition_column_tag = models.Filter(
            must=[models.FieldCondition(key="type", match=models.MatchAny(any=["column_name", "tag"]))]
        )

        filter_condition_glossary = models.Filter(
            must=[models.FieldCondition(key="type", match=models.MatchValue(value="glossary"))]
        )

        task1_coro = self.__vector_search__(
            query_vector=dense_vector,
            using=f"{self.embeddings.model_name}-{EmbeddingsType.DENSE}",
            score_threshold=self.threshold_for_less_relevant_result,
            query_filter=filter_condition_column_tag,
        )
        task2_coro = self.__vector_search__(
            query_vector=late_vector,
            using=f"{self.embeddings.model_name}-{EmbeddingsType.LATE}",
            # score_threshold=self.threshold_for_less_relevant_result,
            query_filter=filter_condition_glossary,
        )

        task1_result, task2_result = await asyncio.gather(task1_coro, task2_coro, return_exceptions=True)

        try:
            if isinstance(task1_result, Exception):
                raise task1_result
            dense_embedding_result = task1_result
            dense_embedding_result = dense_embedding_result[["column_id", "score"]]
        except Exception as ex:
            log.error(ex)
            log.error(f"[*] Error while performing dense search \nReason: {ex}")
            dense_embedding_result = pd.DataFrame()

        try:
            if isinstance(task2_result, Exception):
                raise task2_result
            # Do a late embedding search across glossary
            late_embedding_result_glossary = task2_result
            late_embedding_result_glossary = late_embedding_result_glossary[["column_id", "score"]]
        except Exception as ex:
            log.error(f"[*] Error while performing late search \nReason: {ex}")
            late_embedding_result_glossary = pd.DataFrame()

        if late_embedding_result_glossary.shape[0] > 0:
            # The final score will be total score by no of query embeddings in late vectors
            late_embedding_result_glossary["score"] = late_embedding_result_glossary["score"] / len(late_vector)

            # filter out results that are below designated threshold
            late_embedding_result_glossary = late_embedding_result_glossary.loc[
                late_embedding_result_glossary.score >= self.threshold_for_less_relevant_result
            ].reset_index(drop=True)

        hybrid_result = pd.DataFrame()

        # check if we have results in all cases
        late_shape, dense_shape = (
            late_embedding_result_glossary.shape[0],
            dense_embedding_result.shape[0],
        )

        shapes = [late_shape, dense_shape]

        # Should only merge result if atleast one of the cases of result is present
        if sum(shapes) > 0:
            # filter out results that came out empty
            merging_results = [
                result
                for shape, result in zip(
                    shapes,
                    [
                        late_embedding_result_glossary,
                        dense_embedding_result,
                    ],
                )
                if shape > 0
            ]

            if len(merging_results) == 1:
                # if only one result came then no need to merge

                hybrid_result = merging_results[-1]

            else:
                # merge all non empty results in one go using redude
                hybrid_result = pd.concat(merging_results, axis=0).reset_index(drop=True)[["column_id", "score"]]

        if hybrid_result.shape[0] > 0:
            # Apply max logic to get score
            hybrid_result = hybrid_result.groupby("column_id").agg({"score": "max"}).reset_index()

            if add_relevancy:
                # add the relevancy
                hybrid_result = relevancy_adder(
                    result=hybrid_result,
                    relevancy_scores=self.relevancy_config,
                    i_column="score",
                    o_column="relevancy",
                )

            # hybrid_result.rename(columns={"score": self.hybrid_model}, inplace=True)

        return hybrid_result
    
    async def search(self, query):

        vectors = await self.embeddings.aencode([query], embeddings_types=[EmbeddingsType.DENSE, EmbeddingsType.LATE])

        result = await self.fetch(vectors["dense"][0], vectors["late"][0])

        return result
